###
###  This script was created to test loading larger cubes. This dataset was
###  suggested even though it turned out to just produce 512x512 image planes
###
#from casagui.bokeh.state import initialize_bokeh
#initialize_bokeh( "../../../casalib/dist/esbuild/casalib-v0.0.1.min.js",
#                  "../../../casaguijs/dist/casaguijs.js" )                        ### local build
#initialize_bokeh( "../../../casalib/dist/esbuild/casalib-v0.0.2.min.js" )
#
#initialize_bokeh( bokeh='../../../casaguijs/dist/casaguijs.js' )                   ### local gui/js build + standard bokeh independent library

import os
import sys
import ssl
import certifi
import asyncio
import urllib
import tarfile

from casagui.apps import InteractiveClean

##
## demo measurement set to use
##
ms_path = [ f'2017.1.00661.S/{ms}' for ms in
            [ 'uid___A002_Xc8d560_X66fa_targets_line.ms',
              'uid___A002_Xc91189_X2368_targets_line.ms',
              'uid___A002_Xca525b_X14d6_targets_line.ms',
              'uid___A002_Xca6c94_X4b26_targets_line.ms',
              'uid___A002_Xca6c94_X5376_targets_line.ms',
              'uid___A002_Xcb4a8e_X3734_targets_line.ms',
              'uid___A002_Xcc8b19_X3283_targets_line.ms',
              'uid___A002_Xcc8b19_X7d1b_targets_line.ms' ] ]
##
## where to fetch the demo measurement set
##
ms_url = "https://casa.nrao.edu/download/devel/casavis/data/ms-alma-many-chan-2017_1_00661_S.tar.gz"
##
## output image file name
##
img = 'test'

if not all( map( os.path.isdir, ms_path ) ):
    try:
        response = input('MS is not available, download will be 18GB continue? [Y/N] ')
        while response.lower( ) not in [ 'y', 'n' ]:
            response = input('MS is not available, download will be 18GB continue? [Y/N] ')
        if response.lower( ) == 'y':
            context = ssl.create_default_context(cafile=certifi.where())
            tstream = urllib.request.urlopen(ms_url, context=context, timeout=400)
            tar = tarfile.open(fileobj=tstream, mode="r:gz")
            tar.extractall( )
        else:
            sys.exit("MS not available")
    except urllib.error.URLError:
        print("Failed to open connection to "+ms_url)
        raise

if not all( map( os.path.isdir, ms_path ) ):
    raise  RuntimeError("Failed to fetch measurement set")

###
### https://open-jira.nrao.edu/browse/CAS-13924?focusedCommentId=200087&page=com.atlassian.jira.plugin.system.issuetabpanels%3Acomment-tabpanel#comment-200087
###
### ------------------------------------------------------------------------------------------------------------------------
### tclean(vis=['uid___A002_Xc8d560_X66fa_targets_line.ms', 'uid___A002_Xc91189_X2368_targets_line.ms', 'uid___A002_Xca525b_X14d6_targets_line.ms', 'uid___A002_Xca6c94_X4b26_targets_line.ms', 'uid___A002_Xca6c94_X5376_targets_line.ms', 'uid___A002_Xcb4a8e_X3734_targets_line.ms', 'uid___A002_Xcc8b19_X3283_targets_line.ms', 'uid___A002_Xcc8b19_X7d1b_targets_line.ms'], field='NGC6334I', spw=['16', '16', '16', '16', '16', '16', '16', '16'], antenna=['0,1,2,3,4,5,6,7,8,9,10&', '0,1,2,3,4,5,6,7,8,9,10&', '0,1,2,3,4,5,6,7,8,9&', '0,1,2,3,4,5,6,7,8&', '0,1,2,3,4,5,6,7,8&', '0,1,2,3,4,5,6,7,8,9,10&', '0,1,2,3,4,5,6,7,8,9,10,11&', '0,1,2,3,4,5,6,7,8,9&'], scan=['7,9,12,14,17,19,22,24,27', '7,9,12,14,17,19,22,24,27', '7,9,12,14,17,19,22,24,27', '6,8,11,13,16,18,21,23,26', '6,8,11', '7,9,12,14,17,19,22,24,27', '6,8,11,13,16,18,21,23,26', '6,8,11,13,16,18,21,23,26'], intent='OBSERVE_TARGET#ON_SOURCE', datacolumn='data', imagename='uid___A001_X128e_X222.s38_0.NGC6334I_sci.spw16.cube.I.iter1', imsize=[108, 108], cell=['1.2arcsec'], phasecenter='ICRS 17:20:53.3600 -035.47.00.000', stokes='I', specmode='cube', nchan=2006, start='130.0220940318GHz', width='0.4882638MHz', outframe='LSRK', perchanweightdensity=True, gridder='standard', mosweight=False, usepointing=False, pblimit=0.2, deconvolver='hogbom', restoration=True, restoringbeam='common', pbcor=True, weighting='briggsbwtaper', robust=0.5, npixels=0, niter=99999, threshold='0.292Jy', nsigma=0.0, interactive=0, usemask='auto-multithresh', sidelobethreshold=1.25, noisethreshold=5.0, lownoisethreshold=2.0, negativethreshold=0.0, minbeamfrac=0.1, growiterations=75, dogrowprune=True, minpercentchange=1.0, fastnoise=False, restart=True, savemodel='none', calcres=True, calcpsf=True, parallel=True)
### ------------------------------------------------------------------------------------------------------------------------
###

ic = InteractiveClean( vis=ms_path, imagename=img,
                       field='NGC6334I',
                       spw=['16', '16', '16', '16', '16', '16', '16', '16'],
                       antenna=[ '0,1,2,3,4,5,6,7,8,9,10&', '0,1,2,3,4,5,6,7,8,9,10&', '0,1,2,3,4,5,6,7,8,9&',
                                 '0,1,2,3,4,5,6,7,8&', '0,1,2,3,4,5,6,7,8&', '0,1,2,3,4,5,6,7,8,9,10&',
                                 '0,1,2,3,4,5,6,7,8,9,10,11&', '0,1,2,3,4,5,6,7,8,9&'],
                       scan=['7,9,12,14,17,19,22,24,27', '7,9,12,14,17,19,22,24,27', '7,9,12,14,17,19,22,24,27',
                             '6,8,11,13,16,18,21,23,26', '6,8,11', '7,9,12,14,17,19,22,24,27',
                             '6,8,11,13,16,18,21,23,26', '6,8,11,13,16,18,21,23,26'],
                       intent='OBSERVE_TARGET#ON_SOURCE',
                       datacolumn='data',
                       imsize=[108, 108],
                       cell=['1.2arcsec'],
                       phasecenter='ICRS 17:20:53.3600 -035.47.00.000',
                       stokes='I',
                       specmode='cube',
                       nchan=2006,
                       start='130.0220940318GHz',
                       width='0.4882638MHz',
                       outframe='LSRK',
                       perchanweightdensity=True,
                       gridder='standard',
                       mosweight=False,
                       usepointing=False,
                       pblimit=0.2,
                       deconvolver='hogbom',
                       restoringbeam='common',
                       pbcor=True,
                       weighting='briggsbwtaper',
                       robust=0.5,
                       npixels=0,
                       niter=99999,
                       threshold='0.292Jy',
                       nsigma=0.0,
                       usemask='auto-multithresh',
                       sidelobethreshold=1.25,
                       noisethreshold=5.0,
                       lownoisethreshold=2.0,
                       negativethreshold=0.0,
                       minbeamfrac=0.1,
                       growiterations=75,
                       dogrowprune=True,
                       minpercentchange=1.0,
                       fastnoise=False,
                       savemodel='none',
                       parallel=True
                     )

if True:
    print( "Result: %s" % ic( ) )
    print( " Masks: %s" % repr(ic.masks( )) )
else:
    try:
        asyncio.get_event_loop().run_until_complete(ic.show( ))
        asyncio.get_event_loop().run_forever()

    except KeyboardInterrupt:
        print('\nInterrupt received, shutting down ...')
        #os.system('rm -rf {output_image}.* *.html *.log'.format(output_image=output_image))

    print( "Result: %s" % repr(ic.result( )) )
    print( " Masks: %s" % repr(ic.masks( )) )



from bokeh.layouts import column
from bokeh.models import Button, CustomJS
from bokeh.plotting import show

from svg_icon import SVGIcon

btn = Button(
    icon=SVGIcon(icon_name="stop",
    fill_color="#E1477E",
    spin_duration=2000), 
    label="", 
    button_type="light", 
    default_size=10,
    height_policy="fit", 
    width_policy="fit", 
    margin=(0, 0, 0, 0), 
    sizing_mode="fixed", 
    width=50, 
    height=50 )

btn.js_on_click(CustomJS(code="alert('It works!')"))

show(column(btn))



########################################################################
#
# Copyright (C) 2021,2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Data source for plotting spectra from a CASA image.'''

from bokeh.plotting import ColumnDataSource
from bokeh.util.compiler import TypeScript
from bokeh.core.properties import Instance
from . import ImagePipe
from ..state import casalib_url, casaguijs_url


class SpectraDataSource(ColumnDataSource):
    """Implementation of a ``ColumnDataSource`` customized for spectral lines
    from `CASA`/`CNGI` image cubes. This is designed to use an `ImagePipe` to
    update the spectral line plot displayed in a browser, app or notebook with
    `bokeh`.

    Attributes
    ----------
    image_source: ImagePipe
        the conduit for updating the spectral line data
    """

    image_source = Instance(ImagePipe)

    __javascript__ = [ casalib_url( ), casaguijs_url( ) ]

    def __init__( self, *args, **kwargs ):
        super( ).__init__( *args, **kwargs )
        self.data = self.image_source.spectra( [ 0, 0, 0 ] )



########################################################################
#
# Copyright (C) 2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''implementation of the ``MakeMask`` application for interactive creation
of channel masks'''
import asyncio
from contextlib import asynccontextmanager
from bokeh.layouts import row, column
from bokeh.plotting import show
from bokeh.models import Button, CustomJS
from casagui.toolbox import CubeMask
from casagui.bokeh.components import svg_icon
from bokeh.io import reset_output as reset_bokeh_output
from ..utils import resource_manager, reset_resource_manager

class MakeMask:
    '''Class that can be used to launch a makemask GUI with ``MakeMask('test.image')( )``.'''

    def __stop( self ):
        self.__result_future.set_result(self.__retrieve_result( ))

    def _abort_handler( self, err ):
        self._error_result = err
        self.__stop( )

    def __reset( self ):
        if self.__initialized:
            reset_bokeh_output( )
            reset_resource_manager( )

        ###
        ### reset asyncio result future
        ###
        self.__result_future = None

        ###
        ### used by data pipe (websocket) initialization function
        ###
        self.__initialized = False

        self._cube = CubeMask(self._image_path)
        self._image = None
        ###
        ### error or exception result
        ###
        self._error_result = None

    def __init__( self, image ):
        '''create a ``makemask`` object which will display image planes from a CASA
        image and allow the user to draw masks for each channel.

        Parameters
        ----------
        image: str
            path to CASA image for which interactive masks will be drawn
        '''
        self._image_path = image
        self._cube = CubeMask(self._image_path)
        self._done = None
        self._help = None
        self._help_text = None
        self._layout = None

        ###
        ### This is used to tell whether the websockets have been initialized, but also to
        ### indicate if __call__ is being called multiple times to allow for resetting Bokeh
        ###
        self.__initialized = False

        ###
        ### the asyncio future that is used to transmit the result from interactive clean
        ###
        self.__result_future = None

    def _launch_gui( self ):
        '''create and show GUI
        '''
        self.__initialized = True

        width = 35
        height = 35
        self._done = Button( label="", button_type="danger", max_width=width, max_height=height, name='done',
                             icon=svg_icon(icon_name='makemask-done', size=14) )
        self._help = Button( label="", max_width=width, max_height=height, name='help',
                             icon=svg_icon(icon_name='help', size=14) )

        ### CubeMask provides a help table
        self._help_text = self._cube.help( rows=[ '<tr><td><i>red check button</i></td><td>clicking the red check button will close the dialog and return masks to python</td></tr>' ] )

        self._image = self._cube.image( )
        self._layout = column( row( self._cube.channel_ctrl( ), self._cube.coord_ctrl( ) ),
                               self._image,
                               row( self._cube.slider( ), self._help, self._done ),
                               self._help_text )

        self._done.js_on_click( CustomJS( args=dict( obj=self._cube.js_obj( ) ),
                                          code='''obj.done( )''' ) )                  ### CubeMask should return javascript object which contains
                                                                                      ### a done function to stop cube masking
        self._help.js_on_click( CustomJS( args=dict( help=self._help_text ),
                                          code='''if ( help.visible == true ) help.visible = false
                                                  else help.visible = true''' ) )
        self._cube.connect( )
        show(self._layout)

    def _asyncio_loop( self ):
        '''return the event loop which can be mixed in with an existing event loop
        to allow GUI and websocket events to be processed.
        '''
        return self._cube.loop( )

    def __call__( self ):
        '''Display GUI using the event loop specified by ``loop``.
        '''
        async def _run_( ):
            async with self.serve( ) as s:
                await s[0]
        asyncio.run(_run_( ))
        return self.result( )

    @asynccontextmanager
    async def serve( self ):
        '''This function is intended for developers who would like to embed interactive
        clean as a part of a larger GUI. This embedded use of interactive clean is not
        currently supported and would require the addition of parameters to this function
        as well as changes to the interactive clean implementation. However, this function
        does expose the ``asyncio.Future`` that is used to signal completion of the
        interactive cleaning operation, and it provides the coroutines which must be
        managed by asyncio to make the interactive clean GUI responsive.
        '''
        self.__reset( )
        self._launch_gui( )

        async with self._cube.serve( self.__stop ) as cube:
            self.__result_future = asyncio.Future( )
            yield ( self.__result_future, { 'cube': cube } )

    def __retrieve_result( self ):
        '''If InteractiveClean had a return value, it would be filled in as part of the
        GUI dialog between Python and JavaScript and this function would return it'''
        if isinstance(self._error_result,Exception):
            raise self._error_result
        elif self._error_result is not None:
            return self._error_result
        return self._cube.result( )

    def result( self ):
        '''If InteractiveClean had a return value, it would be filled in as part of the
        GUI dialog between Python and JavaScript and this function would return it'''
        if self.__result_future is None:
            raise RuntimeError( 'no interactive clean result is available' )
        return self.__result_future.result( )



########################################################################
#
# Copyright (C) 2021,2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Functions for coordinate conversion and axis labeling.'''
import numpy as np
from astropy.wcs import WCS
from casatools import image


def get_world_coordinates(ia):
    """
    .. todo::
        Support different world coordiantes
        Provide prettier string formatting

    Get the mapping between pixel coordinates and world coordinates using casatools.coordsys

    Parameters
    -------
    ia: casatools.image

    Returns
    -------
    x_axes_labels, y_axes_labels : (dict, dict)
        Dictionary mapping pixel coordinates to world coordinates
    """
    pix = np.zeros([len(ia.shape()), ia.shape()[0]])
    pix[0, :] = range(ia.shape()[0])
    csys = ia.coordsys()
    world = csys.toworldmany(pix)['numeric']
    x_axes_labels = {i: str(world[0][i]) for i in range(ia.shape()[0])}
    y_axes_labels = {i: str(world[1][i]) for i in range(ia.shape()[0])}
    return x_axes_labels, y_axes_labels


#--- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# This function has not been tested. pylint indicated that 'shp' is not defined. It seems like
# the intent was probably to have the image shape supplied as a parameter since it is not available
# from the coordsys object.
#--- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
def get_world_coordinates_wcs(csys: image.coordsys, shp ) -> list():
    """
    .. todo::
        Support different world coordiantes
        Provide prettier string formatting

    Get the mapping between pixel coordinates and world coordinates using astropy.wcs.WCS

    Parameters
    -------
    csys: casatools.coordsys

    Returns
    -------
    x_axes_labels, y_axes_labels : (dict, dict)
        Dictionary mapping pixel coordinates to world coordinates
    """
    rad_to_deg = 180/np.pi
    w = WCS(naxis=2)
    w.wcs.crpix = csys.referencepixel()['numeric'][0:2]
    w.wcs.cdelt = csys.increment()['numeric'][0:2]*rad_to_deg
    w.wcs.crval = csys.referencevalue()['numeric'][0:2]*rad_to_deg
    w.wcs.ctype = ['RA---SIN', 'DEC--SIN']
    x_axes_labels = {i: w.pixel_to_world(0, i).ra.to_string() for i in range(shp[0] + 1)}
    y_axes_labels = {i: w.pixel_to_world(i, 0).dec.to_string(decimal=True) for i in range(shp[0] + 1)}
    return x_axes_labels, y_axes_labels



import time
import asyncio
import websockets
import traceback

message_count = 0
async def timeserver( websocket ):
    global message_count
    try:
        async for message in websocket:
            message_count += 1
            print(f'{message_count}\t{message}')
            await websocket.send(time.strftime('%Y-%m-%d.%H:%M:%S'))
    except websockets.exceptions.ConnectionClosedError:
        print('browser connection dropped...')
    except Exception:
        print(traceback.print_exc())

async def main( ):
    async with websockets.serve( timeserver, "localhost", 8765 ):
        await asyncio.Future( )

asyncio.run(main( ))



########################################################################
#
# Copyright (C) 2021
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Specialization of the Bokeh ``DataSource`` for image cube data.'''

import numpy as np
from bokeh.plotting import ColumnDataSource
from bokeh.util.compiler import TypeScript
from bokeh.core.properties import Instance, Tuple, Int, Nullable
from bokeh.models.callbacks import Callback
from ._image_pipe import ImagePipe
from ..state import casalib_url, casaguijs_url

class ImageDataSource(ColumnDataSource):
    """Implementation of a ``ColumnDataSource`` customized for planes from
    `CASA`/`CNGI` image cubes. This is designed to use an `ImagePipe` to
    update the image channel/plane displayed in a browser, app or notebook
    with `bokeh`.

    Attributes
    ----------
    image_source: ImagePipe
        the conduit for updating the channel/plane from the image cube
    """

    init_script = Nullable(Instance(Callback), help="""
    JavaScript to be run during initialization of an instance of an ImageDataSource object.
    """)

    image_source = Instance(ImagePipe)
    _mask_contour_source = Nullable(Instance(ColumnDataSource), help='''
    data source for updating contour polygons
    ''')
    num_chans = Tuple( Int, Int, help="[ num-stokes-planes, num-channels ]" )
    cur_chan  = Tuple( Int, Int, help="[ num-stokes-planes, num-channels ]" )

    __javascript__ = [ casalib_url( ), casaguijs_url( ) ]

    def __init__( self, *args, **kwargs ):
        super( ).__init__( *args, **kwargs )
        mask0 = self.image_source.mask0( [0,0] )
        self.data = { 'img': [ self.image_source.channel( [0,0], np.uint8 ) ],
                      'msk0': [ mask0 if mask0 is not None else None ] }
        if self.image_source.have_mask( ):
            self.data['msk'] = [ self.image_source.mask( [0,0] ) ]
        self.num_chans = list(self.image_source.shape[-2:])
        self.cur_chan  = [ 0, 0 ]

    def mask_contour_source( self, data ):
        if not self._mask_contour_source:
            self._mask_contour_source = ColumnDataSource( data=data )
        return self._mask_contour_source

    def pixel_value( self, chan, index ):
        return self.image_source.pixel_value( chan, index )

    def stokes_labels( self ):
        return self.image_source.stokes_labels( )
